@@ -8,7 +8,7 @@ from apscheduler.schedulers.background import BackgroundScheduler
 from websocket import WebSocketApp, WebSocketException
 
 from hiro_graph_client.clientlib import AbstractTokenApiHandler
-from hiro_graph_client.websocketlib import AbstractAuthenticatedWebSocketHandler, ErrorMessage
+from hiro_graph_client.websocketlib import AbstractAuthenticatedWebSocketHandler, ErrorMessage, ReaderStatus
 
 logger = logging.getLogger(__name__)
 """ The logger for this module """
@@ -111,13 +111,17 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
     A handler for issue events
     """
     _events_filter_messages: Dict[str, EventsFilter] = {}
-    _events_filter_messages_lock: threading.RLock
+    _scopes: List[str] = []
+
+    _initial_messages_lock: threading.RLock
+    """Lock for _events_filter_messages and _scopes """
 
     _token_scheduler: BackgroundScheduler
 
     def __init__(self,
                  api_handler: AbstractTokenApiHandler,
                  events_filters: List[EventsFilter],
+                 scopes: List[str] = None,
                  query_params: dict = None):
         """
         Constructor
@@ -125,13 +129,16 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
         :param api_handler: The TokenApiHandler for this WebSocket.
         :param events_filters: Filters for the events. These have to be set or the flood of incoming events will be too
                                big.
+        :param scopes: List of ids of non-default scopes to subscribe. These are ogit/_id of the "ogit/DataScope"s
+                       (i.e. the scope of your instance) you want to subscribe to. Default is None, which means:
+                       Use default scope.
         :param query_params: URL Query parameters for this specific websocket.
         """
         super().__init__(api_handler,
                          'events-ws',
                          query_params=query_params)
 
-        self._events_filter_messages_lock = threading.RLock()
+        self._initial_messages_lock = threading.RLock()
 
         self._token_scheduler = BackgroundScheduler()
 
@@ -144,6 +151,8 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
         for events_filter in events_filters:
             self._events_filter_messages[events_filter.id] = events_filter
 
+        self._scopes = scopes or []
+
     ###############################################################################################################
     # Websocket Events
     ###############################################################################################################
@@ -160,10 +169,14 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
                 self._set_next_token_refresh()
                 self._token_scheduler.start()
 
-            with self._events_filter_messages_lock:
+            with self._initial_messages_lock:
+                for scope in self._scopes:
+                    scope_message = self._get_subscribe_scope_message(scope)
+                    self.send(scope_message)
+
                 for events_filter in self._events_filter_messages.values():
-                    message = self._get_events_register_message(events_filter)
-                    self.send(message)
+                    filter_message = self._get_events_register_message(events_filter)
+                    self.send(filter_message)
 
         except Exception as err:
             raise WebSocketFilterException('Setting events filter failed') from err
@@ -197,6 +210,8 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
             error_message = ErrorMessage.parse(message)
             if error_message:
                 logger.error("Received error: %s", str(error_message))
+                if self._reader_status == ReaderStatus.RUNNING_PRELIMINARY:
+                    self._reader_status = ReaderStatus.FAILED
             else:
                 logger.error("Invalid message: %s", message)
 
@@ -265,10 +280,19 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
 
         return json.dumps(message)
 
+    @staticmethod
+    def _get_subscribe_scope_message(scope_id: str) -> str:
+        message: dict = {
+            "type": "subscribe",
+            "id": scope_id
+        }
+
+        return json.dumps(message)
+
     def add_events_filter(self, events_filter: EventsFilter) -> None:
         message: str = self._get_events_register_message(events_filter)
         self.send(message)
-        with self._events_filter_messages_lock:
+        with self._initial_messages_lock:
             self._events_filter_messages[events_filter.id] = events_filter
 
     def remove_events_filter(self, events_filter_id: str) -> None:
@@ -280,7 +304,7 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
         }
 
         self.send(json.dumps(message))
-        with self._events_filter_messages_lock:
+        with self._initial_messages_lock:
             del self._events_filter_messages[events_filter_id]
 
     def clear_events_filters(self) -> None:
@@ -290,9 +314,23 @@ class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
         }
 
         self.send(json.dumps(message))
-        with self._events_filter_messages_lock:
+        with self._initial_messages_lock:
             self._events_filter_messages = {}
 
+    def subscribe_scope(self, scope_id: str):
+        message = self._get_subscribe_scope_message(scope_id)
+        self.send(json.dumps(message))
+        with self._initial_messages_lock:
+            self._scopes.append(scope_id)
+
+    def remove_scope(self, scope_id: str):
+        """
+        This only removes the scope from the internal list since there is no 'unsubscribe'. You need to
+        restart the websocket for this change to take effect.
+        """
+        with self._initial_messages_lock:
+            self._scopes.remove(scope_id)
+
     ###################################################################################################################
     # Token refresh thread
     ###################################################################################################################
